import datetime
import json
import logging
import os
import shutil
import subprocess
import threading
import time
from glob import glob
from pathlib import Path
from typing import List, Optional

from Pegasus import user
from Pegasus.db import connection
from Pegasus.db.ensembles import Ensembles, Triggers, TriggerType
from Pegasus.db.schema import Trigger


# --- manager ------------------------------------------------------------------
class TriggerManager(threading.Thread):
    """
    Manages workflow triggers. Work is done based on the contents of the
    "trigger" table, and the state of each trigger in that table.
    """

    def __init__(self,):
        threading.Thread.__init__(self, daemon=True)

        self.log = logging.getLogger("trigger.manager")

        # interval in seconds at which the trigger manager will query the
        # database to see if there is work to do (start, stop, restart) triggers
        self.polling_rate = 15

        # references to currently running trigger threads
        # key: (<ensenble_id>, <trigger_name>), value: handle to trigger thread
        self.running = dict()

        self.trigger_dao = None
        self.ensemble_dao = None

    def run(self):
        """Trigger manager main loop."""

        self.log.info("trigger manager starting")

        while True:
            # TODO: user will always be the same.., keep this in the loop or move it out
            u = user.get_user_by_uid(os.getuid())
            session = connection.connect(
                u.get_master_db_url(), connect_args={"check_same_thread": False}
            )

            try:
                self.trigger_dao = Triggers(session)
                self.ensemble_dao = Ensembles(session)
                triggers = self.trigger_dao.list_triggers()
                self.log.info("processing {} triggers".format(len(triggers)))
                for t in triggers:
                    t_name = TriggerManager.get_tname(t)
                    if t.state == "READY":
                        self.start_trigger(t)
                    elif t.state == "RUNNING" and t_name not in self.running:
                        # restart
                        self.log.debug("{} not in memory, restarting it".format(t_name))
                        self.start_trigger(t)
                    elif t.state == "RUNNING" and not self.running[t_name].is_alive():
                        # exited
                        self.log.debug(
                            "{} exited, removing references to it".format(t_name)
                        )
                        self.stop_trigger(t)
                    elif t.state == "STOPPED":
                        self.stop_trigger(t)
            finally:
                session.close()

            time.sleep(self.polling_rate)

    def start_trigger(self, trigger: Trigger):
        """Given a trigger, start the appropriate trigger thread.

        :param trigger: the trigger to be started
        :type trigger: Trigger
        """

        trigger_name = TriggerManager.get_tname(trigger)
        self.log.debug("starting {}".format(trigger_name))

        workflow = json.loads(trigger.workflow)
        required_args = {
            "ensemble_id": trigger.ensemble_id,
            "ensemble": self.ensemble_dao.get_ensemble_name(trigger.ensemble_id),
            "trigger": trigger.name,
            "workflow_script": workflow["script"],
            "workflow_args": workflow["args"] if workflow["args"] else [],
        }
        trigger_specific_kwargs = json.loads(trigger.args)

        # create trigger thread
        if trigger._type == TriggerType.CRON.value:
            t = CronTrigger(**required_args, **trigger_specific_kwargs)

        elif trigger._type == TriggerType.FILE_PATTERN.value:
            t = FilePatternTrigger(**required_args, **trigger_specific_kwargs)

        else:
            raise NotImplementedError(
                "unsupported trigger type: {}".format(trigger.type)
            )

        # keep ref to trigger thread
        self.running[trigger_name] = t
        t.start()

        # update state
        self.log.debug(
            "changing {name} state: {old_state} -> {new_state}".format(
                name=trigger_name, old_state=trigger.state, new_state="RUNNING",
            )
        )
        self.trigger_dao.update_state(
            ensemble_id=trigger.ensemble_id, trigger_id=trigger._id, new_state="RUNNING"
        )

    def stop_trigger(self, trigger: Trigger):
        """Stop a trigger thread.

        :param trigger: the trigger to be stopped
        :type trigger: Trigger
        """
        # using reference to trigger, tell trigger to shutdown
        target_trigger = TriggerManager.get_tname(trigger)
        self.log.debug("stopping {}".format(target_trigger))
        self.running[target_trigger].shutdown()
        del self.running[target_trigger]

        # remove entry from database
        self.trigger_dao.delete_trigger(trigger.ensemble_id, trigger.name)

    @staticmethod
    def get_tname(trigger: Trigger) -> tuple:
        """Given a trigger object, get its name as a pair (<ensemble_id>, <trigger_name>)

        :param trigger: the trigger
        :type trigger: Trigger
        :return: pair (<ensemble_id>, <trigger_name>)
        :rtype: tuple
        """
        return (trigger.ensemble_id, trigger.name)


# trigger threads --------------------------------------------------------------
class TriggerThread(threading.Thread):
    """Base class for trigger thread implementations."""

    def __init__(
        self,
        ensemble_id: int,
        ensemble: str,
        trigger: str,
        workflow_script: str,
        workflow_args: List[str] = [],
    ):

        threading.Thread.__init__(self, name=(ensemble_id, trigger))

        self.log = logging.getLogger("trigger.{}::{}".format(ensemble_id, trigger))
        self.ensemble_id = ensemble_id
        self.ensemble = ensemble
        self.trigger = trigger
        self.workflow_cmd = [workflow_script]
        if workflow_args:
            self.workflow_cmd.extend(workflow_args)

        # thread stopping condition
        self.stop_event = threading.Event()

    def shutdown(self):
        """Gracefully shutdown this thread by setting the stop_event."""
        self.log.info("shutting down".format(self.name))
        self.stop_event.set()


class CronTrigger(TriggerThread):
    """Submits a workflow to the ensemble manager at a specified interval."""

    def __init__(
        self,
        ensemble_id: int,
        ensemble: str,
        trigger: str,
        interval: int,
        workflow_script: str,
        workflow_args: Optional[List[str]] = None,
        timeout: Optional[int] = None,
        **kwargs
    ):

        TriggerThread.__init__(
            self,
            ensemble_id=ensemble_id,
            ensemble=ensemble,
            trigger=trigger,
            workflow_script=workflow_script,
            workflow_args=workflow_args,
        )

        self.timeout = int(timeout) if timeout else None
        self.interval = int(interval)
        self.elapsed = 0

    def __repr__(self):
        return "<CronTrigger {} interval={}s>".format(self.name, self.interval)

    def run(self):
        """CronTrigger main loop."""
        try:
            self.log.debug("starting")

            while not self.stop_event.isSet():
                cmd = [
                    "pegasus-em",
                    "submit",
                    "{ens}.{tr}_{ts}".format(
                        ens=self.ensemble,
                        tr=self.trigger,
                        ts=int(datetime.datetime.now().timestamp()),
                    ),
                ]
                cmd.extend(self.workflow_cmd)

                cp = subprocess.run(cmd, stderr=subprocess.PIPE)

                if cp.returncode == 0:
                    self.log.info("executed cmd: {}".format(cmd))
                else:
                    self.log.error("encountered an error executing cmd: {}".format(cmd))
                    stderr = (
                        cp.stderr.decode()
                        if isinstance(cp.stderr, bytes)
                        else cp.stderr
                    )
                    raise RuntimeError(stderr)

                time.sleep(self.interval)
                if self.timeout:
                    self.elapsed += self.interval

                    if self.elapsed >= self.timeout:
                        self.log.debug("timed out")
                        break

        except Exception:
            self.log.exception("error")
        finally:
            self.log.debug("exited")


class FilePatternTrigger(TriggerThread):
    """
    Collects files based on the given filepatterns and submits workflow to the
    ensemble manager at a specified interval.
    """

    def __init__(
        self,
        ensemble_id: str,
        ensemble: str,
        trigger: str,
        interval: int,
        workflow_script: str,
        workflow_args: Optional[List[str]] = None,
        timeout: Optional[int] = None,
        **kwargs
    ):
        TriggerThread.__init__(
            self,
            ensemble_id=ensemble_id,
            ensemble=ensemble,
            trigger=trigger,
            workflow_script=workflow_script,
            workflow_args=workflow_args,
        )

        self.timeout = int(timeout) if timeout else None
        self.interval = int(interval)
        self.elapsed = 0
        self.file_patterns = kwargs["file_patterns"]

    def __repr__(self):
        return "<FilePatternTrigger {} interval={}s patterns={}>".format(
            self.name, self.interval, self.file_patterns
        )

    def run(self):
        """FilePatternTrigger main loop."""
        try:
            self.log.debug("starting")

            while not self.stop_event.isSet():

                files = self.collect_and_move_files()

                if len(files) > 0:
                    cmd = [
                        "pegasus-em",
                        "submit",
                        "{ens}.{tr}_{ts}".format(
                            ens=self.ensemble,
                            tr=self.trigger,
                            ts=int(datetime.datetime.now().timestamp()),
                        ),
                    ]
                    cmd.extend(self.workflow_cmd)
                    cmd.append("--inputs")
                    cmd.extend(files)

                    cp = subprocess.run(cmd, stderr=subprocess.PIPE)

                    if cp.returncode == 0:
                        self.log.info("executed cmd: {}".format(cmd))
                    else:
                        self.log.error(
                            "encountered error executing cmd: {}".format(cmd)
                        )
                        stderr = (
                            cp.stderr.decode()
                            if isinstance(cp.stderr, bytes)
                            else cp.stderr
                        )
                        raise RuntimeError(stderr)

                time.sleep(self.interval)
                if self.timeout:
                    self.elapsed += self.interval

                    if self.elapsed >= self.timeout:
                        self.log.debug("timed out")
                        break

        except Exception:
            self.log.exception("error")
        finally:
            self.log.debug("exited")

    def collect_and_move_files(self) -> List[str]:
        """
        Collect absolute paths of all files that match the given file
        patterns, then move those files into a newly created subdirectory
        meant for previously processed files. For example, if the patterns
        ["/inputs/*.txt", "/inputs2/*.txt] are given, and "/inputs/f1.txt" and
        "/inputs2/f2.txt" exist, the result will be that those two paths are returned
        AND those two files will be moved to "/inputs/processed" and "/inputs2/processed"
        respectively. This is to avoid looping over an increasingly larger set
        of files as more inputs arrive.

        :return: list of paths to matched files
        :rtype: List[str]
        """
        collected = []
        for pattern in self.file_patterns:
            # create dir to move files to if it doesn't already exist
            parent_dir = Path(pattern).parent.resolve()
            processed_dir = parent_dir / "processed"
            processed_dir.mkdir(parents=False, exist_ok=True)

            for match in glob(pattern):
                # move file into processed dir and add that path
                f = Path(match).resolve()
                dst = processed_dir / f.name

                shutil.move(str(f), str(dst))
                collected.append(str(dst))

        return collected



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__author__ = "Rafael Ferreira da Silva"

DB_VERSION = 1

import logging

from sqlalchemy.exc import *

from Pegasus.db.admin.admin_loader import *
from Pegasus.db.admin.versions.base_version import *

log = logging.getLogger(__name__)


class Version(BaseVersion):
    def __init__(self, connection):
        super().__init__(connection)

    def update(self, force=False):
        log.info("Updating to version %s" % DB_VERSION)
        try:
            self.db.execute("SELECT site FROM rc_lfn LIMIT 0,1")
            return
        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            self.db.rollback()
            raise DBAdminError(e)

        try:
            log.debug("  Updating database schema...")
            log.debug("    Creating new table...")
            self.db.execute("CREATE TABLE rc_lfn_new ( LIKE rc_lfn )")
            self.db.execute("ALTER TABLE rc_lfn_new ADD COLUMN site VARCHAR(245)")
            log.debug("    Removing index...")
            self.db.execute("ALTER TABLE rc_lfn_new DROP INDEX sk_rc_lfn")
            log.debug("    Adding new constraint...")
            self.db.execute(
                "ALTER TABLE rc_lfn_new ADD CONSTRAINT sk_rc_lfn UNIQUE(lfn,pfn,site)"
            )
            log.debug("    Copying data...")
            self.db.execute("INSERT INTO rc_lfn_new(id, lfn, pfn) SELECT * FROM rc_lfn")

            log.debug("    Renaming table...")
            if self.db.get_bind().driver == "mysqldb":
                self.db.execute(
                    "RENAME TABLE rc_lfn TO rc_lfn_old, rc_lfn_new TO rc_lfn"
                )
            else:
                self.db.execute("ALTER TABLE rc_lfn RENAME TO rc_lfn_old")
                self.db.execute("ALTER TABLE rc_lfn_new RENAME TO rc_lfn")

            log.debug("    Droping old table...")
            self.db.execute("ALTER TABLE rc_attr DROP FOREIGN KEY fk_rc_attr")
            self.db.execute("DROP TABLE rc_lfn_old")
            log.debug("  Data schema successfully updated.")

            log.debug("  Migrating attribute data...")
            self.db.execute(
                "UPDATE rc_lfn l INNER JOIN rc_attr a ON (l.id=a.id AND a.name='pool') SET l.site=a.value"
            )
            log.debug("  Migration successfully completed.")

            log.debug("  Cleaning the database...")
            self.db.execute("DELETE FROM rc_attr WHERE name='pool'")
            log.debug("  Database successfully cleaned.")

            log.debug("  Adding new foreign constraint")
            self.db.execute(
                "ALTER TABLE rc_attr ADD CONSTRAINT fk_rc_attr FOREIGN KEY (id) REFERENCES rc_lfn(id)"
            )
            log.debug("  Foreign constraint successfully added.")

            log.debug("  Validating the update process...")
            data = self.db.execute(
                "SELECT COUNT(id) FROM rc_attr WHERE name='pool'"
            ).first()
            if data is not None:
                if data[0] > 0:
                    self.db.rollback()
                    raise DBAdminError(
                        "attribute pool failed to be removed. There are still %d entries."
                        % data[0]
                    )

            updated = self.db.execute(
                "SELECT COUNT(id) FROM rc_lfn WHERE site IS NOT NULL"
            ).first()
            log.debug("  Updated %d entries in the database." % updated)
            self.db.commit()

        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            self.db.rollback()
            raise DBAdminError(e)

    def downgrade(self, force=False):
        try:
            data = self.db.execute(
                "SELECT COUNT(id) AS c FROM rc_Lfn GROUP BY lfn, pfn ORDER BY c DESC LIMIT 1"
            ).first()
            if data is not None:
                count = int(data[0])
                if count > 1 and not force:
                    raise DBAdminError(
                        "A possible data loss was detected: use '--force' to ignore this message."
                    )

            log.debug("  Updating database schema...")
            log.debug("    Creating new table...")
            self.db.execute("CREATE TABLE rc_lfn_new ( LIKE rc_lfn )")
            self.db.execute("ALTER TABLE rc_lfn_new DROP INDEX sk_rc_lfn")
            self.db.execute("ALTER TABLE rc_lfn_new DROP COLUMN site")
            log.debug("    Adding new constraint...")
            self.db.execute(
                "ALTER TABLE rc_lfn_new ADD CONSTRAINT sk_rc_lfn UNIQUE(lfn,pfn)"
            )
            log.debug("    Copying data...")
            self.db.execute(
                "INSERT INTO rc_lfn_new(id, lfn, pfn) SELECT MAX(id), lfn, pfn FROM rc_lfn GROUP BY lfn, pfn"
            )
            self.db.execute(
                "INSERT INTO rc_attr(id, name, value) SELECT n.id, 'pool', site FROM rc_lfn o INNER JOIN rc_lfn_new n ON (n.id = o.id)"
            )
            self.db.execute(
                "DELETE FROM rc_attr WHERE NOT EXISTS (SELECT id FROM rc_lfn_new WHERE rc_lfn_new.id = rc_attr.id)"
            )
            log.debug("    Renaming table...")
            self.db.execute("RENAME TABLE rc_lfn TO rc_lfn_old, rc_lfn_new TO rc_lfn")
            log.debug("    Droping old table...")
            self.db.execute("ALTER TABLE rc_attr DROP FOREIGN KEY fk_rc_attr")
            self.db.execute("DROP TABLE rc_lfn_old")
            self.db.execute(
                "ALTER TABLE rc_attr ADD CONSTRAINT fk_rc_attr FOREIGN KEY (id) REFERENCES rc_lfn(id)"
            )
            log.debug("  Data schema successfully updated.")
            self.db.commit()
        except OperationalError:
            pass
        except Exception as e:
            self.db.rollback()
            raise DBAdminError(e)



import atexit
import datetime
import json
import os
import re
import sys
import uuid
from optparse import OptionParser

from Pegasus.cluster import RecordParser
from Pegasus.monitoring.metadata import Metadata
from Pegasus.tools import kickstart_parser

# logging
log = {
    "name": None,
    "timestamp": None,
    "exitcode": None,
    "app_exitcode": None,
    "retry": None,
}
tmp_log_files = []


class JobFailed(Exception):
    pass


def rotate_file(outfile, errfile):
    """Rename .out and .err files to .out.XXX and .err.XXX where XXX
    is the next sequence number. Returns the new name, or fails with
    an error message and a non-zero exit code."""

    # This is just to prevent the file from being accidentally renamed
    # again in testing.
    if re.search(r"\.out\.[0-9]{3}$", outfile):
        return outfile, errfile

    # Must end in .out
    if not outfile.endswith(".out"):
        raise JobFailed("%s does not look like a kickstart .out file" % outfile)

    # Find next file in sequence
    retry = None
    for i in range(0, 1000):
        candidate = "%s.%03d" % (outfile, i)
        if not os.path.isfile(candidate):
            retry = i
            break

    # unlikely to occur
    if retry is None:
        raise JobFailed("%s has been renamed too many times!" % (outfile))

    basename = outfile[:-4]
    log["retry"] = retry

    # rename .out to .out.000
    newout = "%s.out.%03d" % (basename, retry)
    os.rename(outfile, newout)

    # rename .err to .err.000 if it exists
    newerr = None
    if os.path.isfile(errfile):
        newerr = "%s.err.%03d" % (basename, retry)
        os.rename(errfile, newerr)

    return newout, newerr


def readfile(filename):
    # If the file does not exits, return empty string
    if filename is None or not os.path.isfile(filename):
        return ""

    f = open(filename)
    try:
        return f.read()
    finally:
        f.close()


def find_cluster_summary(txt):
    "Find and return the cluster summary record if it exists"

    b = txt.find("[cluster-summary")
    if b < 0:
        # Older seqexec used this name
        b = txt.find("[seqexec-summary")
    if b >= 0:
        e = txt.find("]", b)
        if e < 0:
            raise JobFailed("Invalid cluster-summary record")

        return RecordParser(txt[b : e + 1]).parse()

    # If we found a cluster-task, but no cluster-summary, then there is a problem
    b = txt.find("[cluster-task")
    if b >= 0:
        raise JobFailed("cluster-summary is missing")

    return None


def check_cluster_summary(record):
    "Make sure that the cluster-summary record is OK"

    if "stat" in record:
        stat = record["stat"]
        if stat != "ok":
            raise JobFailed("cluster-summary stat=%s" % stat)

    # If any of the tasks failed, then job failed
    if "failed" in record:
        failed = int(record["failed"])
        if failed > 0:
            raise JobFailed("cluster-summary failed=%d" % failed)

    # If no tasks were submitted, then it succeeded
    if "submitted" in record:
        submitted = int(record["submitted"])
        if submitted == 0:
            return 0

    # If there were no tasks, then the job was successful
    if "tasks" in record:
        tasks = int(record["tasks"])
        if tasks == 0:
            return 0

    # If the number of successes was zero, then the job failed
    if "succeeded" in record:
        succeeded = int(record["succeeded"])
        if succeeded == 0:
            raise JobFailed("cluster-summary succeeded=%d" % succeeded)

    return 0


def check_kickstart_records(txt):
    # Check the exitcodes of all kickstart records
    regex = re.compile(r'raw="(-?[0-9]+)"')
    succeeded = 0
    e = 0

    # first check for integrity errors - there are rare cases where most
    # of the jobs look successful, but kickstart was unable to calculate
    # a checksum for an output file
    if re.search(r"integrity_error: ", txt):
        raise JobFailed("detected integrity creation errors in the job")

    # yaml
    for m in re.finditer(r"raw: ([0-9]+)", txt):
        raw = int(m.group(1))
        log["app_exitcode"] = raw
        if raw != 0:
            raise JobFailed("task exited with raw status %d" % raw)
        succeeded = succeeded + 1

    # xml
    while True:
        b = txt.find("<status", e)
        if b < 0:
            break
        e = txt.find("</status>", b)
        if e < 0:
            raise JobFailed("mismatched <status>")
        e = e + len("</status>")
        m = regex.search(txt[b:e])
        if m:
            raw = int(m.group(1))
            log["app_exitcode"] = raw
        else:
            raise JobFailed("<status> was missing valid 'raw' attribute")
        if raw != 0:
            raise JobFailed("task exited with raw status %d" % raw)
        succeeded = succeeded + 1

    # Fail if there were no invocation records and no cluster-summary
    if succeeded == 0:
        raise JobFailed("No successful kickstart records")

    return 0


def unquote_message(message):
    def genchars():
        yield from message

    chars = genchars()
    output = []
    for c in chars:
        if c == "+":
            output.append(" ")
        elif c == "\\":
            try:
                c = next(chars)
            except StopIteration:
                output.append(c)
                break
            if c == "+":
                output.append("+")
            else:
                output.append("\\")
                output.append(c)
        else:
            output.append(c)

    return "".join(output)


def unquote_messages(messages):
    return [unquote_message(m) for m in messages]


def has_any_failure_messages(stdio, messages):
    """Return true if any of the messages appear in the files"""
    if len(messages) == 0:
        return False

    messages = unquote_messages(messages)

    for txt in stdio:
        for m in messages:
            if m in txt:
                return True

    return False


def has_all_success_messages(stdio, messages):
    """Return true if any of the messages don't appear in the files"""
    if len(messages) == 0:
        return True

    messages = unquote_messages(messages)

    found = set()
    for txt in stdio:
        for m in messages:
            if m in txt:
                found.add(m)

    for m in messages:
        if m not in found:
            return False

    return True


def get_errfile(outfile):
    """Get the stderr file name given the stdout file name"""
    i = outfile.rfind(".out")
    left = outfile[0:i]
    right = ""
    if i + 5 < len(outfile):
        right = outfile[i + 4 :]
    errfile = left + ".err" + right
    return errfile


def append_to_wf_metadata_log(files_metadata, logfile):
    """
    Writes out file metadata to a logfile in the File RC format

    :param files_metadata:    list of FileMetadata objects
    :param logfile:           the log file to append the info to
    :return:
    """
    # writing to log file (concurrency safe)
    with open(logfile, "a", encoding="utf8") as outfile:
        for file_metadata in files_metadata:
            res = file_metadata.convert_to_rce()
            outfile.write(res + "\n")


def exitcode(
    outfile,
    check_invocations=True,
    dagman_job_status=None,
    rename=True,
    failure_messages=[],
    success_messages=[],
    wf_metadata_log=None,
    generate_meta=True,
):
    if not os.path.isfile(outfile):
        raise JobFailed("%s does not exist" % outfile)

    errfile = get_errfile(outfile)

    # outfile Must end in .out
    if not outfile.endswith(".out"):
        raise JobFailed("%s does not look like a kickstart .out file" % outfile)

    meta_file = outfile[:-3] + "meta"

    # If we are renaming, then rename
    if rename:
        outfile, errfile = rotate_file(outfile, errfile)

    # First, check exitcode supplied by DAGMan, if any
    if dagman_job_status is not None:
        log["app_exitcode"] = dagman_job_status
        if dagman_job_status != 0:
            # if dagman has flagged the job as a failure, then always flag the job as error
            # PM-1746 highlights the case where jobs that get held are aborted eventually
            # by dagman
            raise JobFailed("dagman reported non-zero exitcode: %d" % dagman_job_status)

    # Next, read the output and error files
    stdout = readfile(outfile)
    stderr = readfile(errfile)

    # Next, check the size of the output file
    # when a job is launched without kickstart stdout can be empty.
    if check_invocations and len(stdout) == 0:
        raise JobFailed("Empty stdout")

    # Next, if we have failure messages, then fail if we find one in the
    # output of the job
    if has_any_failure_messages([stdout, stderr], failure_messages):
        raise JobFailed("Failure message found in output")

    # Next, if we have success messages, then fail if we don't find all
    # in the output of the job
    if not has_all_success_messages([stdout, stderr], success_messages):
        raise JobFailed("Success message missing from output")

    # Next, check exitcodes of all tasks
    cs = find_cluster_summary(stdout)
    if cs is not None:
        check_cluster_summary(cs)
    else:
        # PM-927 , PM-1821 Only check kickstart records if -c is specified
        if check_invocations:
            check_kickstart_records(stdout)

    # Next check if metadata file needs to be generated
    if generate_meta:
        files_metadata = parse_metadata_from_kickstart(outfile)
        # always generate a meta file even if it is a zero byte file
        directory = os.path.dirname(meta_file)
        basename = os.path.basename(meta_file)
        Metadata.write_to_jsonfile(
            files_metadata, directory, basename, prefix="pegasus-exitcode"
        )

        if wf_metadata_log and files_metadata:
            # PM-1257 write files metadata to workflow log
            append_to_wf_metadata_log(files_metadata, wf_metadata_log)


def parse_metadata_from_kickstart(outfile):
    # First assume we will find rotated file
    parser = kickstart_parser.Parser(outfile)
    kickstart_output = parser.parse_stampede()

    # Start empty
    files = []
    for record in kickstart_output:
        if "invocation" in record:
            # Ok, we have an invocation record, extract the metadata information
            if "outputs" in record:
                for lfn in record["outputs"].keys():
                    files.append(record["outputs"][lfn])

    return files


def _log_info(info_msg):
    if len(tmp_log_files) > 0:
        tmp_log_files[0].write(info_msg + "\n")
    else:
        print(info_msg)


def _log_error(err_msg):
    if len(tmp_log_files) > 0:
        tmp_log_files[1].write(err_msg + "\n")
    else:
        print(err_msg)


def _write_logs(log_filename):
    if log_filename:
        # reading std_out and std_err files
        std_out = tmp_log_files[0]
        std_err = tmp_log_files[1]
        std_out.close()
        std_err.close()

        with open(std_out.name) as sout:
            log["std_out"] = sout.read()
        with open(std_err.name) as serr:
            log["std_err"] = serr.read()

        # writing to log file (concurrency safe)
        with open(log_filename, "a", encoding="utf8") as outfile:
            res = json.dumps(log, ensure_ascii=False)
            outfile.write(res + "\n")

    else:
        print(json.dumps(log))


def main(args):
    usage = "Usage: %prog [options] job.out"
    parser = OptionParser(usage)

    parser.add_option(
        "-r",
        "--return",
        action="store",
        type="int",
        dest="status",
        metavar="R",
        help="Return code reported by DAGMan. This can be specified in a "
        "DAG using the $RETURN variable.",
    )
    parser.add_option(
        "-n",
        "--no-rename",
        action="store_false",
        dest="rename",
        default=True,
        help="Don't rename job.out and .err to .out.XXX and .err.XXX. "
        "Useful for testing.",
    )
    parser.add_option(
        "-N",
        "--no-metadata",
        action="store_false",
        dest="generate_meta",
        default=True,
        help="disable generation of metadata file after parsing of kickstart records",
    )
    parser.add_option(
        "-I",
        "--no-invocations",
        action="store_false",
        dest="check_invocations",
        default=True,
        help="do not check for invocation records(present in kickstart output) output in the job.out file",
    )
    parser.add_option(
        "-f",
        "--failure-message",
        action="append",
        dest="failure_messages",
        default=[],
        help="Failure message to find in job stdout/stderr. If this "
        "message exists in the stdout/stderr of the job, then the "
        "job will be considered a failure no matter what other "
        "output exists. If multiple failure messages are provided, "
        "then none of them can exist in the output or the job is "
        "considered a failure.",
    )
    parser.add_option(
        "-s",
        "--success-message",
        action="append",
        dest="success_messages",
        default=[],
        help="Success message to find in job stdout/stderr. If this "
        "message does not exist in the stdout/stderr of the job, "
        "then the job will be considered a failure no matter what "
        "other output exists. If multiple success messages are "
        "provided, then they must all exist in the output or "
        "the job is considered a failure.",
    )
    parser.add_option(
        "-l",
        "--log",
        action="store",
        type="string",
        dest="log_filename",
        help="Name of the common log file in which stdout/stderr will" "be redirected.",
    )
    parser.add_option(
        "-M",
        "--metadata-log",
        action="store",
        type="string",
        dest="wf_metadata_log",
        help="Name of the common log file in which the metadata parsed "
        "from .out file is placed in append mode for the workflow",
    )

    (options, args) = parser.parse_args(args)

    if len(args) != 1:
        parser.error("Please specify job.out")

    outfile = args[0]

    if options.log_filename:
        # temporary log files
        tmp_log_name = "_exit-code-" + str(uuid.uuid4())
        tmp_log_files.append(open(tmp_log_name + ".out", "w"))
        tmp_log_files.append(open(tmp_log_name + ".err", "w"))

    try:
        log["name"] = outfile
        log["timestamp"] = datetime.datetime.now().isoformat()
        exitcode(
            outfile,
            check_invocations=options.check_invocations,
            dagman_job_status=options.status,
            rename=options.rename,
            failure_messages=options.failure_messages,
            success_messages=options.success_messages,
            generate_meta=options.generate_meta,
            wf_metadata_log=options.wf_metadata_log,
        )
        log["exitcode"] = 0
        _write_logs(options.log_filename)
        sys.exit(0)
    except JobFailed as jf:
        _log_error(str(jf))
        log["exitcode"] = 1
        _write_logs(options.log_filename)
        sys.exit(1)


@atexit.register
def remove_temporary_files():
    # cleaning temporary files
    for tmp_file in tmp_log_files:
        os.remove(tmp_file.name)



#!/usr/bin/env python3


import argparse
import os
import sys
from glob import glob
from os.path import abspath, dirname, exists, join, normpath


def _perl_dump(**kw):
    """."""
    print(
        """my $pegasus_bin_dir = "%(bin_dir)s";
my $pegasus_conf_dir = "%(conf_dir)s";
my $pegasus_java_dir = "%(java_dir)s";
my $pegasus_perl_dir = "%(perl_dir)s";
my $pegasus_python_dir = "%(python_dir)s";
my $pegasus_python_externals_dir = "%(python_externals_dir)s";
my $pegasus_share_dir = "%(share_dir)s";
my $pegasus_schema_dir = "%(schema_dir)s";
unshift(@INC, $pegasus_perl_dir);
"""
        % kw,
        end="",
    )


def _perl_hash(**kw):
    """."""
    print(
        """use vars qw(%%pegasus);
%%pegasus =
    ( bin => "%(conf_dir)s"
    , conf => "%(conf_dir)s"
    , java => "%(java_dir)s"
    , perl => "%(perl_dir)s"
    , python => "%(python_dir)s"
    , pyexts => "%(python_externals_dir)s"
    , share => "%(share_dir)s"
    , schema => "%(schema_dir)s"
    );
unshift( @INC, $pegasus{perl} );
"""
        % kw,
        end="",
    )


def _python_hash(**kw):
    """."""
    print(
        """pegasus_bin_dir = "%(bin_dir)s"
pegasus_conf_dir = "%(conf_dir)s"
pegasus_java_dir = "%(java_dir)s"
pegasus_perl_dir = "%(perl_dir)s"
pegasus_python_dir = "%(python_dir)s"
pegasus_python_externals_dir = "%(python_externals_dir)s"
pegasus_share_dir = "%(share_dir)s"
pegasus_schema_dir = "%(schema_dir)s"
"""
        % kw,
        end="",
    )


def _sh_dump(**kw):
    """."""
    print(
        """PEGASUS_BIN_DIR="%(bin_dir)s";
export PEGASUS_BIN_DIR
PEGASUS_CONF_DIR="%(conf_dir)s"
export PEGASUS_CONF_DIR
PEGASUS_JAVA_DIR="%(java_dir)s"
export PEGASUS_JAVA_DIR
PEGASUS_PERL_DIR="%(perl_dir)s"
export PEGASUS_PERL_DIR
PEGASUS_PYTHON_DIR="%(python_dir)s"
export PEGASUS_PYTHON_DIR
PEGASUS_PYTHON_EXTERNALS_DIR="%(python_externals_dir)s"
export PEGASUS_PYTHON_EXTERNALS_DIR
PEGASUS_SHARE_DIR="%(share_dir)s"
export PEGASUS_SHARE_DIR
PEGASUS_SCHEMA_DIR="%(schema_dir)s"
export PEGASUS_SCHEMA_DIR
CLASSPATH="%(classpath)s"
export CLASSPATH
"""
        % kw,
        end="",
    )


def _get_bin_dir(exe):
    bin_dir = normpath(join(dirname(abspath(exe)), "bin"))

    while not exists(bin_dir):
        bin_dir = normpath(join(bin_dir, "..", "..", "bin"))

    return bin_dir


def _main(
    version=False,
    perl_dump=False,
    perl_hash=False,
    python_hash=False,
    sh_dump=False,
    bin=False,
    conf=False,
    java=False,
    perl=False,
    python=False,
    python_externals=False,
    schema=False,
    r=False,
    classpath=False,
    noeoln=False,
):
    """."""
    _version = "@PEGASUS_VERSION@"

    bin_dir = _get_bin_dir(sys.argv[0])
    base_dir = dirname(bin_dir)

    lib = "@LIBDIR@"  # lib64 for 64bit RPMS
    if lib.startswith("@"):
        lib = "lib"

    python_lib = "@PYTHON_LIBDIR@"
    if python_lib.startswith("@"):
        python_lib = "lib/pegasus/python"

    conf_dir = join(base_dir, "etc")
    share_dir = join(base_dir, "share", "pegasus")
    java_dir = join(share_dir, "java")
    perl_dir = join(base_dir, lib, "pegasus", "perl")
    python_dir = join(base_dir, python_lib)
    python_externals_dir = join(base_dir, lib, "pegasus", "externals", "python")
    schema_dir = join(share_dir, "schema")
    r_dir = "".join(sorted(glob(join(share_dir, "r", "*.tar.gz"))))

    # for development - running out of a source checkout
    test = join(base_dir, "build", "classes")
    extra_classpath = test if exists(test) else ""

    # in native packaging mode, some directories move
    if base_dir == "/usr":
        conf_dir = "/etc/pegasus"

    # classpath
    jars = sorted(glob(join(java_dir, "*.jar")))
    if extra_classpath:
        jars.insert(0, extra_classpath)

    _classpath = ":".join(jars)
    if "CLASSPATH" in os.environ:
        _classpath += ":" + os.environ["CLASSPATH"]

    # construct aws batch classpath
    aws_jars = sorted(glob(join(java_dir, "aws", "*.jar")))
    _classpath += ":" + ":".join(aws_jars)

    eol = "" if noeoln else "\n"

    if version:
        print(_version, end=eol)
    elif perl_dump:
        _perl_dump(
            bin_dir=bin_dir,
            conf_dir=conf_dir,
            java_dir=java_dir,
            perl_dir=perl_dir,
            python_dir=python_dir,
            python_externals_dir=python_externals_dir,
            share_dir=share_dir,
            schema_dir=schema_dir,
        )
    elif perl_hash:
        _perl_hash(
            bin_dir=bin_dir,
            conf_dir=conf_dir,
            java_dir=java_dir,
            perl_dir=perl_dir,
            python_dir=python_dir,
            python_externals_dir=python_externals_dir,
            share_dir=share_dir,
            schema_dir=schema_dir,
        )
    elif python_hash:
        _python_hash(
            bin_dir=bin_dir,
            conf_dir=conf_dir,
            java_dir=java_dir,
            perl_dir=perl_dir,
            python_dir=python_dir,
            python_externals_dir=python_externals_dir,
            share_dir=share_dir,
            schema_dir=schema_dir,
        )
    elif sh_dump:
        _sh_dump(
            bin_dir=bin_dir,
            conf_dir=conf_dir,
            java_dir=java_dir,
            perl_dir=perl_dir,
            python_dir=python_dir,
            python_externals_dir=python_externals_dir,
            share_dir=share_dir,
            schema_dir=schema_dir,
            classpath=_classpath,
        )
    elif bin:
        print(bin_dir, end=eol)
    elif conf:
        print(conf_dir, end=eol)
    elif java:
        print(java_dir, end=eol)
    elif perl:
        print(perl_dir, end=eol)
    elif python:
        print(python_dir, end=eol)
    elif python_externals:
        print(python_externals_dir, end=eol)
    elif schema:
        print(schema_dir, end=eol)
    elif r:
        print(r_dir, end=eol)
    elif classpath:
        print(_classpath, end=eol)
    else:
        # Code should not reach here.
        pass


def main():
    """."""
    parser = argparse.ArgumentParser(
        description="This is NOT an application to configure Pegasus, but an application to query the current Pegasus installation."
    )

    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        help="Print Pegasus version information and exit.",
    )

    parser.add_argument(
        "--perl-dump",
        action="store_true",
        help="Dumps all settings in perl format as separate variables.",
    )
    parser.add_argument(
        "--perl-hash",
        action="store_true",
        help="Dumps all settings in perl format as single perl hash.",
    )
    parser.add_argument(
        "--python-hash",
        action="store_true",
        help="Dumps all settings in python format.",
    )
    parser.add_argument(
        "--sh-dump", action="store_true", help="Dumps all settings in shell format.",
    )

    parser.add_argument(
        "--bin",
        action="store_true",
        help="Print the directory containing Pegasus binaries.",
    )
    parser.add_argument(
        "--conf",
        action="store_true",
        help="Print the directory containing configuration files.",
    )
    parser.add_argument(
        "--java", action="store_true", help="Print the directory containing the jars.",
    )
    parser.add_argument(
        "--perl",
        action="store_true",
        help="Print the directory to include into your PERL5LIB.",
    )
    parser.add_argument(
        "--python",
        action="store_true",
        help="Print the directory to include into your PYTHONPATH.",
    )
    parser.add_argument(
        "--python-externals",
        action="store_true",
        help="Print the directory to the external Python libraries.",
    )
    parser.add_argument(
        "--schema", action="store_true", help="Print the directory containing schemas.",
    )
    parser.add_argument(
        "--r",
        action="store_true",
        help="Print the path to the R DAX API source package.",
    )
    parser.add_argument(
        "--classpath",
        action="store_true",
        help="Builds a classpath containing the Pegasus jars.",
    )

    parser.add_argument(
        "--noeoln",
        action="store_true",
        help="Do not produce a end-of-line after output. This is useful when being called from non-shell backticks in scripts.",
    )

    # Ensure at least one option is passed
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()
    _main(**vars(args))


if __name__ == "__main__":
    main()



import ast
import logging

from sqlalchemy import and_, not_, or_
from sqlalchemy.sql.operators import ColumnOperators as operator

COMPARE_OPERATORS = {
    ast.Eq: operator.__eq__,
    ast.NotEq: operator.__ne__,
    ast.Lt: operator.__lt__,
    ast.LtE: operator.__le__,
    ast.Gt: operator.__gt__,
    ast.GtE: operator.__ge__,
    ast.Is: operator.is_,
    ast.IsNot: operator.isnot,
    ast.In: operator.in_,
    ast.NotIn: operator.notin_,
}

OPERATORS = {
    ast.Add: operator.__add__,
    ast.Sub: operator.__sub__,
    ast.Mult: operator.__mul__,
    ast.Div: operator.__div__,
}

BOOLEAN_OPERATORS = {ast.And: and_, ast.Or: or_}

UNARY_OPERATORS = {
    ast.Not: not_,
    # ast.UAdd: operator.pos,
    # ast.USub: operator.neg
}

BUILTINS = {"None": None}


class InvalidQueryError(Exception):
    pass


def query_parse(clause, **symbols):
    """
    Parses query clause.

    Query Clause Identifier Rules:
        Prefix + Identifier

    Prefix Rules:
        Is optional
        Must start with a-z OR A-Z
        Can contain a-z OR A-Z OR 0-9 OR _
        Must end with .

    Identifier Rules:
        Is mandatory
        Must start with a-z OR A-Z
        Can contain a-z OR A-Z OR 0-9 OR _

    Supported Comparators (case-insensitive)
        - ==
        - !=
        - <
        - <=
        - >
        - >=
        - in

    Supported Functions
        - like
        - ilike

    Supported Operands (case-insensitive)
        - and
        - or
        - not

    Null Handling

    <identifier> is None OR <identifier> == None

    Examples:

        - age > 1
        - user.age >= 10 and user.age < 100
        - ((user.age < 18 and movie.ratings in ("PG", "PG13") or user.age > 18)

    :param clause: The clause to be parsed.
    :type clause: str

    :raises InvalidQueryError: Raised if the `clause` is not valid.

    :returns: sqlalchemy boolean clause list.
    :rtype: tuple

    .. example::

        >>> str( query_parse("w.wf_id == 1") )
        'workflow.wf_id = :wf_id_1'

        >>> str( query_parse("w.wf_id > 1 and w.wf_id < 5") )
        'workflow.wf_id > :wf_id_1 AND workflow.wf_id < :wf_id_2'
    """

    try:
        n = ast.parse(clause.strip(",") + ",", mode="eval").body
    except SyntaxError as e:
        raise InvalidQueryError("Invalid query: %s" % e)

    if not isinstance(n, ast.Tuple):
        raise InvalidQueryError("Invalid condition: must evaluate to a boolean value")

    return _QueryEvaluator(**symbols).visit(n)


class _QueryEvaluator(ast.NodeVisitor):
    def __init__(self, **symbols):
        super().__init__()
        self._symbols = symbols
        self._log = logging.getLogger(__name__)

    # Compare
    def visit_Compare(self, n):
        # Compare(expr left, cmpop* ops, expr* comparators)
        left = self.visit(n.left)

        result = True
        for i, v in zip(n.ops, n.comparators):
            op = COMPARE_OPERATORS[i.__class__]
            comparator = self.visit(v)

            result = result and op(left, comparator)
            left = comparator

        return result

    # BoolOp
    def visit_BoolOp(self, n):
        # boolop = And | Or
        op = BOOLEAN_OPERATORS[n.op.__class__]
        return op(*(self.visit(i) for i in n.values))

    # BinOp
    def visit_BinOp(self, n):
        left = self.visit(n.left)
        right = self.visit(n.right)
        op = OPERATORS[n.op.__class__]
        return op(left, right)

    # UnaryOp
    def visit_UnaryOp(self, n):
        # UnaryOp(unaryop op, expr operand)
        self._log.info("UnaryOp <%s> <%s>", n.op.__class__, n.operand)
        op = UNARY_OPERATORS[n.op.__class__]
        operand = self.visit(n.operand)
        return op(operand)

    # Identifiers
    def visit_Attribute(self, n):
        self._log.info("Attribute <%s> <%s>", n.value, n.attr)
        value = self.visit(n.value)
        if isinstance(value, dict):
            return value[n.attr]
        return getattr(value, n.attr)

    def visit_Subscript(self, n):
        self._log.info("Subscript <%s> <%s>", n.value, n.slice)
        value = self.visit(n.value)
        slice = self.visit(n.slice)
        return value[slice]

    def visit_Index(self, n):
        return self.visit(n.value)

    def visit_keyword(self, n):
        return {n.arg: n.value}

    def visit_Name(self, n):
        self._log.info("Name <%s>", n.id)
        name = self._symbols.get(n.id, BUILTINS.get(n.id, None))
        if n.id not in self._symbols and n.id not in BUILTINS:
            raise NameError(
                "Invalid name <%s> at Line <%d> Col "
                "<%d>" % (n.id, n.lineno, n.col_offset)
            )

        return name

    # Literals
    def visit_Num(self, n):
        return n.n

    def visit_Str(self, n):
        return n.s

    def visit_NameConstant(self, n):
        self._log.info("NameConstant <%s>", n.value)
        return n.value

    def visit_List(self, n):
        return [self.visit(n) for n in n.elts]

    def visit_Tuple(self, n):
        return tuple(self.visit(e) for e in n.elts)

    def visit_Set(self, n):
        return {self.visit(n) for n in n.elts}

    def visit_Call(self, n):
        args = []
        try:
            func = self.visit(n.func)
        except AttributeError:
            raise InvalidQueryError(
                "Invalid name <%s> at Line <%d> Col <%d>"
                % (n.func.attr, n.lineno, n.col_offset)
            )

        # Args
        args.extend([self.visit(a) for a in n.args])

        # Keyword Args
        kwargs = {}
        for k in n.keywords:
            kwargs.update(self.visit(k))

        return func(*args, **kwargs)

    # Catch All
    def generic_visit(self, n):
        raise InvalidQueryError(
            "Invalid query at Line <%d> Col <%d>" % (n.lineno, n.col_offset)
        )


def _main():
    import sys

    logging.basicConfig(level=logging.DEBUG)
    logging.debug("Expression <%s>", sys.argv[1])
    from Pegasus.db.schema import Workflow

    result = query_parse(sys.argv[1], Workflow)
    logging.debug("Evaluation Result <%s>", result[0])


if __name__ == "__main__":
    _main()



#  Copyright 2007-2014 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__author__ = "Rajiv Mayani"

import hashlib
import logging

from sqlalchemy.orm import aliased, defer, joinedload
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import and_, desc, distinct, func

from Pegasus.db import connection
from Pegasus.db.admin.admin_loader import DBAdminError
from Pegasus.db.errors import StampedeDBNotFoundError
from Pegasus.db.schema import (
    RCLFN,
    RCPFN,
    Host,
    Invocation,
    Job,
    JobInstance,
    Jobstate,
    MasterWorkflow,
    MasterWorkflowstate,
    RCMeta,
    Task,
    TaskMeta,
    Workflow,
    WorkflowFiles,
    WorkflowMeta,
    Workflowstate,
)
from Pegasus.service import cache
from Pegasus.service._query import InvalidQueryError, query_parse
from Pegasus.service._sort import InvalidSortError, sort_parse
from Pegasus.service.base import PagedResponse

log = logging.getLogger(__name__)


class WorkflowQueries:
    def __init__(self, connection_string, use_cache=True):
        if connection_string is None:
            raise ValueError("Connection string is required")

        self._conn_string_csum = hashlib.md5(
            connection_string.encode("utf-8")
        ).hexdigest()

        try:
            self.session = connection.connect(connection_string)
        except (connection.ConnectionError, DBAdminError) as e:
            log.exception(e)
            raise StampedeDBNotFoundError

        self._use_cache = True
        self.use_cache = use_cache

    def close(self):
        self.session.close()

    @property
    def use_cache(self):
        return self._use_cache

    @use_cache.setter
    def use_cache(self, use_cache):
        if isinstance(use_cache, bool):
            self._use_cache = use_cache
        else:
            raise TypeError("Expecting boolean, found %s" % type(use_cache))

    def _cache_key_from_query(self, q):
        statement = q.with_labels().statement
        compiled = statement.compile()
        params = compiled.params

        cache_key = " ".join(
            [self._conn_string_csum, str(compiled)]
            + [str(params[k]) for k in sorted(params)]
        )
        return hashlib.md5(cache_key.encode("utf-8")).hexdigest()

    def _get_count(self, q, use_cache=True, timeout=60):
        cache_key = "%s.count" % self._cache_key_from_query(q)
        if use_cache and cache.get(cache_key):
            log.debug("Cache Hit: %s" % cache_key)
            count = cache.get(cache_key)

        else:
            log.debug("Cache Miss: %s" % cache_key)
            count = q.count()
            t = timeout(count) if callable(timeout) else timeout
            cache.set(cache_key, count, t)

        return count

    def _get_all(self, q, use_cache=True, timeout=60):
        cache_key = "%s.all" % self._cache_key_from_query(q)
        if use_cache and cache.get(cache_key):
            log.debug("Cache Hit: %s" % cache_key)
            record = cache.get(cache_key)

        else:
            log.debug("Cache Miss: %s" % cache_key)
            record = q.all()
            t = timeout(record) if callable(timeout) else timeout
            cache.set(cache_key, record, t)

        return record

    def _get_one(self, q, use_cache=True, timeout=60):
        cache_key = "%s.one" % self._cache_key_from_query(q)
        if use_cache and cache.get(cache_key):
            log.debug("Cache Hit: %s" % cache_key)
            record = cache.get(cache_key)

        else:
            log.debug("Cache Miss: %s" % cache_key)
            record = q.one()
            t = timeout(record) if callable(timeout) else timeout
            cache.set(cache_key, record, t)

        return record

    @staticmethod
    def _evaluate_query(q, query, **resource):
        if not query:
            return q

        try:
            x = query_parse(query, **resource)
            q = q.filter(x[0])
        except (KeyError, AttributeError):
            log.exception("Invalid query %s" % query)
            raise InvalidQueryError("Invalid query %s" % query)
        except NameError as e:
            log.exception("Invalid field %s" % e)
            raise InvalidQueryError("Invalid field %s" % e)
        except Exception:
            log.exception("Invalid query %s" % query)
            raise InvalidQueryError("Invalid query %s" % query)

        return q

    @staticmethod
    def _add_ordering(q, order, **resource):
        if not q or not order or not resource:
            return q

        sort_order = sort_parse(order)

        for prefix, identifier, sort_dir in sort_order:
            try:
                field = getattr(resource[prefix], identifier)
            except (KeyError, AttributeError):
                log.exception("Invalid field {}.{}".format(prefix, identifier))
                raise InvalidSortError("Invalid field %r" % identifier)

            if sort_dir == "ASC":
                q = q.order_by(field)
            else:
                q = q.order_by(desc(field))

        return q

    @staticmethod
    def _add_pagination(q, start_index=None, max_results=None, total_records=None):
        """
        LIMIT <skip>, <count>       - Valid
        LIMIT <count> OFFSET <skip> - Valid
        OFFSET <skip>               - Invalid

        If only start_index is provided and total_records is known then we can compute both limit and offset to
        effectively support OFFSET <skip>
        """
        if start_index and max_results:
            q = q.offset(start_index)
            q = q.limit(max_results)

        elif not start_index and not max_results:
            return q

        else:
            if max_results:
                q = q.limit(max_results)

            elif total_records:
                q = q.offset(start_index)
                q = q.limit(total_records)

        return q


class MasterWorkflowQueries(WorkflowQueries):
    def get_root_workflows(
        self,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Root Workflow objects.

        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Collection of tuples (MasterWorkflow, MasterWorkflowstate)
        """

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(MasterWorkflow)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            log.debug("total_records 0")
            return PagedResponse([], 0, 0)

        #
        # Finish Construction of Base SQLAlchemy Query `q`
        #
        qws = self._get_max_master_workflow_state()
        qws = qws.subquery("master_workflowstate")

        alias = aliased(MasterWorkflowstate, qws)

        q = q.outerjoin(qws, MasterWorkflow.wf_id == qws.c.wf_id)
        q = q.add_entity(alias)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, r=MasterWorkflow, ws=alias)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, r=MasterWorkflow)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        for i in range(len(records)):
            new_record = records[i][0]
            new_record.workflow_state = records[i][1]
            new_record.__includes__ = ("workflow_state",)
            records[i] = new_record

        return PagedResponse(records, total_records, total_filtered)

    def get_root_workflow(self, m_wf_id, use_cache=True):
        """
        Returns a Root Workflow object identified by m_wf_id.

        :param m_wf_id: m_wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param use_cache: If available, use cached results

        :return: Root Workflow object
        """
        q = self.session.query(MasterWorkflow)

        if m_wf_id is None:
            raise ValueError("m_wf_id cannot be None")

        m_wf_id = str(m_wf_id)
        if m_wf_id.isdigit():
            q = q.filter(MasterWorkflow.wf_id == m_wf_id)
        else:
            q = q.filter(MasterWorkflow.wf_uuid == m_wf_id)

        #
        # Finish Construction of Base SQLAlchemy Query `q`
        #
        qws = self._get_max_master_workflow_state(m_wf_id=m_wf_id)
        qws = qws.subquery("master_workflowstate")

        q = q.outerjoin(qws, MasterWorkflow.wf_id == qws.c.wf_id)
        q = q.add_entity(aliased(MasterWorkflowstate, qws))

        try:
            record_tuple = self._get_one(q, use_cache)
            record = record_tuple[0]
            record.workflow_state = record_tuple[1]
            return record

        except NoResultFound as e:
            log.exception("Not Found: Root Workflow for given m_wf_id (%s)" % m_wf_id)
            raise e

    def _get_max_master_workflow_state(self, m_wf_id=None, mws=MasterWorkflowstate):
        qmax = self._get_recent_master_workflow_state(m_wf_id, mws)
        qmax = qmax.subquery("max_timestamp")

        q = self.session.query(mws)
        q = q.join(
            qmax, and_(mws.wf_id == qmax.c.wf_id, mws.timestamp == qmax.c.max_time)
        )

        return q

    def _get_recent_master_workflow_state(self, m_wf_id=None, mws=MasterWorkflowstate):
        q = self.session.query(mws.wf_id)
        q = q.add_columns(func.max(mws.timestamp).label("max_time"))

        if m_wf_id:
            log.debug("filter on m_wf_id")
            q = q.filter(mws.wf_id == m_wf_id)

        q = q.group_by(mws.wf_id)

        return q


class StampedeWorkflowQueries(WorkflowQueries):
    def wf_uuid_to_wf_id(self, wf_id):
        if wf_id is None:
            raise ValueError("wf_id cannot be None")

        wf_id = str(wf_id)
        if not wf_id.isdigit():
            q = self.session.query(Workflow.wf_id)
            q = q.filter(Workflow.wf_uuid == wf_id)

            try:
                q = self._get_one(q, True, timeout=600)
                wf_id = q.wf_id

            except NoResultFound as e:
                raise e

        return wf_id

    # Workflow

    def get_workflows(
        self,
        m_wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=False,
        **kwargs
    ):
        """
        Returns a collection of the Workflow objects.

        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Collection of Workflow objects
        """
        m_wf_id = self.wf_uuid_to_wf_id(m_wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Workflow)
        q = q.filter(Workflow.root_wf_id == m_wf_id)
        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, w=Workflow)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, w=Workflow)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_workflow(self, wf_id, use_cache=True):
        """
        Returns a Workflow object identified by wf_id.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid

        :return: Workflow object
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        q = self.session.query(Workflow)
        q = q.filter(Workflow.wf_id == wf_id)

        try:
            return self._get_one(q, use_cache)
        except NoResultFound as e:
            raise e

    # Workflow Meta

    def get_workflow_meta(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=False,
        **kwargs
    ):
        """
        Returns a collection of the Workflowstate objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Workflow Meta collection, total records count, total filtered records count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(WorkflowMeta)
        q = q.filter(WorkflowMeta.wf_id == wf_id)
        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, wm=WorkflowMeta)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, wm=WorkflowMeta)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    # Workflow Files

    def get_workflow_files(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=False,
        **kwargs
    ):
        """
        Returns a collection of all files associated with the Workflow.

        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Collection of Workflow Files
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(WorkflowFiles)
        q = q.filter(WorkflowFiles.wf_id == wf_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        q = (
            self.session.query(WorkflowFiles)
            .options(
                joinedload(WorkflowFiles.lfn).joinedload(RCLFN.pfns),
                joinedload(WorkflowFiles.lfn).joinedload(RCLFN.meta),
            )
            .filter(WorkflowFiles.wf_id == wf_id)
        )

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, l=RCLFN, p=RCPFN, rm=RCMeta)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, l=RCLFN, p=RCPFN, rm=RCMeta)

        records = self._get_all(q, use_cache)

        _records = []
        for r in records:
            o = {k: getattr(r, k) for k in r.__table__.columns.keys()}
            o["lfn"] = r.lfn.lfn
            o["pfns"] = r.lfn.pfns
            o["meta"] = r.lfn.meta
            _records.append(o)

        return PagedResponse(_records, total_records, total_filtered)

    # Workflow State

    def get_workflow_state(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Workflowstate objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Workflow States collection, total records count, total filtered records count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        # Use shorter caching timeout
        timeout = 5

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Workflowstate)
        q = q.filter(Workflowstate.wf_id == wf_id)

        total_records = total_filtered = self._get_count(q, use_cache, timeout=timeout)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, ws=Workflowstate)
            total_filtered = self._get_count(q, use_cache, timeout=timeout)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, ws=Workflowstate)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache, timeout=timeout)

        return PagedResponse(records, total_records, total_filtered)

    def _get_max_workflow_state(self, wf_id=None, ws=Workflowstate):
        qmax = self._get_recent_workflow_state(wf_id, ws)
        qmax = qmax.subquery("max_timestamp")

        q = self.session.query(ws)
        q = q.join(
            qmax, and_(ws.wf_id == qmax.c.wf_id, ws.timestamp == qmax.c.max_time)
        )

        return q

    def _get_recent_workflow_state(self, wf_id=None, ws=Workflowstate):
        q = self.session.query(ws.wf_id)
        q = q.add_columns(func.max(ws.timestamp).label("max_time"))

        if wf_id:
            log.debug("filter on wf_id")
            q = q.filter(ws.wf_id == wf_id)

        q = q.group_by(ws.wf_id)

        return q

    # Job

    def get_workflow_jobs(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Job objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Jobs collection, total jobs count, filtered jobs count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Job)
        q = q.filter(Job.wf_id == wf_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, j=Job)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, j=Job)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_job(self, job_id, use_cache=True):
        """
        Returns a Job object identified by job_id.

        :param job_id: ID of the job
        :param use_cache: If available, use cached results

        :return: job record
        """
        if job_id is None:
            raise ValueError("job_id cannot be None")

        q = self.session.query(Job)
        q = q.filter(Job.job_id == job_id)

        try:
            return self._get_one(q, use_cache, timeout=600)
        except NoResultFound as e:
            raise e

    # Host

    def get_workflow_hosts(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Host objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: hosts collection, total jobs count, filtered jobs count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Host)
        q = q.filter(Host.wf_id == wf_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, h=Host)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, h=Host)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_host(self, host_id, use_cache=True):
        """
        Returns a Host object identified by host_id.

        :param host_id: Id of the host
        :param use_cache: If available, use cached results

        :return: host record
        """
        if host_id is None or not str(host_id).isdigit():
            raise ValueError("host_id cannot be None")

        q = self.session.query(Host)
        q = q.filter(Host.host_id == host_id)

        try:
            return self._get_one(q, use_cache)
        except NoResultFound as e:
            raise e

    # Job State

    def get_job_instance_states(
        self,
        wf_id,
        job_id,
        job_instance_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the JobInstanceState objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param job_id: job_id associated with the job instance states
        :param job_instance_id: job_instance_id associated with the job instance states
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: state record
        """
        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Jobstate)
        q = q.join(JobInstance, JobInstance.job_instance_id == Jobstate.job_instance_id)
        q = q.join(Job, Job.job_id == JobInstance.job_id)

        q = q.filter(Job.wf_id == wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)
        q = q.filter(Jobstate.job_instance_id == job_instance_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, js=Jobstate)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, js=Jobstate)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def _get_recent_job_state(self, job_instance_id=None, js=Jobstate):
        q = self.session.query(js.job_instance_id)
        q = q.add_columns(func.max(js.jobstate_submit_seq).label("max_jsss"))

        if job_instance_id:
            log.debug("filter on job_instance_id")
            q = q.filter(js.job_instance_id == job_instance_id)

        q = q.group_by(js.job_instance_id)

        return q

    # Task

    def get_workflow_tasks(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Task objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Collection of Task objects
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Task)
        q = q.filter(Task.wf_id == wf_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, t=Task)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, t=Task)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_job_tasks(
        self,
        wf_id,
        job_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Task objects.

        :param job_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Collection of Task objects
        """
        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Task)
        q = q.filter(Task.wf_id == wf_id)
        q = q.filter(Task.job_id == job_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, t=Task)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, t=Task)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_task(self, task_id, use_cache=True):
        """
        Returns a Task object identified by task_id.

        :param task_id: Id of the task
        :param use_cache: If available, use cached results

        :return: task record
        """
        q = self.session.query(Task)

        if task_id is None:
            raise ValueError("task_id cannot be None")

        q = q.filter(Task.task_id == task_id)

        try:
            return self._get_one(q, use_cache)
        except NoResultFound as e:
            raise e

    # Task Meta

    def get_task_meta(
        self,
        task_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=False,
        **kwargs
    ):
        """
        Returns a collection of the TaskMeta objects.

        :param task_id: Id of the task
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Workflow Meta collection, total records count, total filtered records count
        """
        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(TaskMeta)
        q = q.filter(TaskMeta.task_id == task_id)
        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, tm=TaskMeta)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, tm=TaskMeta)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    # Job Instance

    def get_job_instances(
        self,
        wf_id,
        job_id,
        recent=False,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the JobInstance objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param job_id: job_id associated with the job instances
        :param recent: Get the most recent results
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: job-instance collection, total jobs count, filtered jobs count
        """
        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(JobInstance)
        q = q.join(Job, Job.job_id == JobInstance.job_id)

        q = q.filter(Job.wf_id == wf_id)
        q = q.filter(Job.job_id == job_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        if recent:
            qjss = self._get_recent_job_instance(job_id)
            qjss = qjss.subquery("max_jss")
            q = q.join(
                qjss,
                and_(
                    JobInstance.job_id == qjss.c.job_id,
                    JobInstance.job_submit_seq == qjss.c.max_jss,
                ),
            )

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query or recent:
            q = self._evaluate_query(q, query, ji=JobInstance)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, ji=JobInstance)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_job_instance(self, job_instance_id, use_cache=True, timeout=5):
        """
        Returns a JobInstance object identified by job_instance_id.

        :param job_instance_id: Id of the job instance
        :param use_cache: If available, use cached results
        :param timeout: Duration for which the job-instance should be cached, if exitcode is None i.e. Job is running

        :return: job-instance record
        """

        def timeout_duration(ji):
            return 300 if ji and ji.exitcode is not None else timeout

        if job_instance_id is None:
            raise ValueError("job_instance_id cannot be None")

        q = self.session.query(JobInstance)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)

        try:
            return self._get_one(q, use_cache, timeout=timeout_duration)
        except NoResultFound as e:
            raise e

    def _get_recent_job_instance(self, job_id=None, ji=JobInstance):
        q = self.session.query(ji.job_id)
        q = q.add_columns(func.max(ji.job_submit_seq).label("max_jss"))

        if job_id:
            log.debug("filter on job_id")
            q = q.filter(ji.job_id == job_id)

        q = q.group_by(ji.job_id)

        return q

    # Invocation

    def get_workflow_invocations(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Invocation objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: invocations record
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Invocation)
        q = q.filter(Invocation.wf_id == wf_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, i=Invocation)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, i=Invocation)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_job_instance_invocations(
        self,
        wf_id,
        job_id,
        job_instance_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the Invocation objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param job_id: Id of the job associated with the invocation
        :param job_instance_id: Id of the job instance associated with the invocation
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: invocations record
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Invocation)
        q = q.filter(Invocation.wf_id == wf_id)
        q = q.filter(Invocation.job_instance_id == job_instance_id)

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, i=Invocation)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, i=Invocation)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)

        return PagedResponse(records, total_records, total_filtered)

    def get_invocation(self, invocation_id, use_cache=True):
        """
        Returns a Invocation object identified by invocation_id.

        :param invocation_id: Id of the invocation
        :param use_cache: If available, use cached results

        :return: invocation record
        """
        if invocation_id is None or not str(invocation_id).isdigit():
            raise ValueError("invocation_id cannot be None")

        q = self.session.query(Invocation)
        q = q.filter(Invocation.invocation_id == invocation_id)

        try:
            return self._get_one(q, use_cache)
        except NoResultFound as e:
            raise e

    # Views

    def get_running_jobs(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the running Job objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Jobs collection, total jobs count, filtered jobs count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Job, JobInstance).options(
            defer(JobInstance.stdout_text), defer(JobInstance.stderr_text)
        )
        q = q.filter(Job.job_id == JobInstance.job_id)

        q = q.filter(Job.wf_id == wf_id)
        q = q.filter(JobInstance.exitcode == None)  # noqa: E711

        # Recent
        qjss = self._get_recent_job_instance()
        qjss = qjss.filter(JobInstance.exitcode != None)  # noqa: E711
        qjss = qjss.subquery("max_jss")

        q = q.join(
            qjss,
            and_(
                JobInstance.job_id == qjss.c.job_id,
                JobInstance.job_submit_seq == qjss.c.max_jss,
            ),
        )

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, j=Job, ji=JobInstance)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, j=Job, ji=JobInstance)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)
        records = self._merge_job_instance(records)

        return PagedResponse(records, total_records, total_filtered)

    def get_successful_jobs(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the successful Job objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Jobs collection, total jobs count, filtered jobs count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Job, JobInstance).options(
            defer(JobInstance.stdout_text), defer(JobInstance.stderr_text)
        )
        q = q.filter(Job.job_id == JobInstance.job_id)

        q = q.filter(Job.wf_id == wf_id)
        q = q.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode == 0
        )  # noqa: E711

        # Recent
        qjss = self._get_recent_job_instance()
        qjss = qjss.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode == 0
        )  # noqa: E711
        qjss = qjss.subquery("max_jss")

        q = q.join(
            qjss,
            and_(
                JobInstance.job_id == qjss.c.job_id,
                JobInstance.job_submit_seq == qjss.c.max_jss,
            ),
        )

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, j=Job, ji=JobInstance)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, j=Job, ji=JobInstance)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)
        records = self._merge_job_instance(records)

        return PagedResponse(records, total_records, total_filtered)

    def get_failed_jobs(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the failed Job objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Jobs collection, total jobs count, filtered jobs count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Job, JobInstance).options(
            defer(JobInstance.stdout_text), defer(JobInstance.stderr_text)
        )
        q = q.filter(Job.job_id == JobInstance.job_id)

        q = q.filter(Job.wf_id == wf_id)
        q = q.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode != 0
        )  # noqa: E711

        # Recent
        qjss = self._get_recent_job_instance()
        qjss = qjss.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode != 0
        )  # noqa: E711
        qjss = qjss.subquery("max_jss")

        q = q.join(
            qjss,
            and_(
                JobInstance.job_id == qjss.c.job_id,
                JobInstance.job_submit_seq == qjss.c.max_jss,
            ),
        )

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, j=Job, ji=JobInstance)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, j=Job, ji=JobInstance)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)
        records = self._merge_job_instance(records)

        return PagedResponse(records, total_records, total_filtered)

    def get_failing_jobs(
        self,
        wf_id,
        start_index=None,
        max_results=None,
        query=None,
        order=None,
        use_cache=True,
        **kwargs
    ):
        """
        Returns a collection of the failing Job objects.

        :param wf_id: wf_id is wf_id iff it consists only of digits, otherwise it is wf_uuid
        :param start_index: Return results starting from record `start_index`
        :param max_results: Return a maximum of `max_results` records
        :param query: Filtering criteria
        :param order: Sorting criteria
        :param use_cache: If available, use cached results

        :return: Jobs collection, total jobs count, filtered jobs count
        """
        wf_id = self.wf_uuid_to_wf_id(wf_id)

        #
        # Construct SQLAlchemy Query `q` to count.
        #
        q = self.session.query(Job, JobInstance).options(
            defer(JobInstance.stdout_text), defer(JobInstance.stderr_text)
        )

        q = q.filter(Job.wf_id == wf_id)
        q = q.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode != 0
        )  # noqa: E711

        q = q.filter(Job.job_id == JobInstance.job_id)

        # Running
        j = aliased(Job, name="j")
        ji = aliased(JobInstance, name="ji")

        qr = self.session.query(distinct(j.job_id))

        qr = qr.filter(j.wf_id == wf_id)
        qr = qr.filter(ji.exitcode == None)  # noqa: E711

        qr = qr.filter(j.job_id == ji.job_id)
        qr = qr.subquery()

        q = q.filter(Job.job_id.in_(qr))

        # Recent
        qjss = self._get_recent_job_instance()
        qjss = qjss.filter(Job.wf_id == wf_id)
        qjss = qjss.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode != 0
        )  # noqa: E711

        qjss = qjss.filter(Job.job_id == JobInstance.job_id)
        qjss = qjss.subquery("allmaxjss")

        q = q.filter(
            and_(
                JobInstance.job_id == qjss.c.job_id,
                JobInstance.job_submit_seq == qjss.c.max_jss,
            )
        )

        total_records = total_filtered = self._get_count(q, use_cache)

        if total_records == 0:
            return PagedResponse([], 0, 0)

        #
        # Construct SQLAlchemy Query `q` to filter.
        #
        if query:
            q = self._evaluate_query(q, query, j=Job, ji=JobInstance)
            total_filtered = self._get_count(q, use_cache)

            if total_filtered == 0 or (start_index and start_index >= total_filtered):
                log.debug("total_filtered is 0 or start_index >= total_filtered")
                return PagedResponse([], total_records, total_filtered)

        #
        # Construct SQLAlchemy Query `q` to sort
        #
        if order:
            q = self._add_ordering(q, order, j=Job, ji=JobInstance)

        #
        # Construct SQLAlchemy Query `q` to paginate.
        #
        q = WorkflowQueries._add_pagination(q, start_index, max_results, total_filtered)

        records = self._get_all(q, use_cache)
        records = self._merge_job_instance(records)

        return PagedResponse(records, total_records, total_filtered)

    @staticmethod
    def _merge_job_instance(records):
        if records:
            for i in range(len(records)):
                new_record = records[i][0]
                new_record.job_instance = records[i][1]
                new_record.__includes__ = ("job_instance",)
                records[i] = new_record

        return records
